--!A cross-platform build utility based on Lua
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--
-- Copyright (C) 2015-present, Xmake Open Source Community.
--
-- @author      ruki
-- @file        repo.lua
--

task("repo")
    set_category("plugin")
    on_run("main")
    set_menu {
        usage = "xmake repo [options] [name] [url] [branch]",
        description = "Manage package repositories.",
        options = {
            {'a', "add",    "k", nil, "Add the given remote repository url."},
            {'r', "remove", "k", nil, "Remove the given remote repository url."},
            {'u', "update", "k", nil, "Update all local repositories from the remote."},
            {'c', "clear",  "k", nil, "Clear all added repositories."},
            {},
            {'l', "list",   "k", nil, "List all added repositories."},
            {'g', "global", "k", nil, "Save repository to global. (default: local)"},
            {},
            {nil, "name",   "v", nil, "The repository name."},
            {nil, "url",    "v", nil, "The repository url"},
            {nil, "branch", "v", nil, "The repository branch"}
        }
    }
