//
//  main.m
//  test3
//
//  Created by ruki on 2020/4/4.
//  Copyright © 2020 tboox. All rights reserved.
//

#import <Cocoa/Cocoa.h>

int main(int argc, const char * argv[]) {
    @autoreleasepool {
        // Setup code that might create autoreleased objects goes here.
    }
    return NSApplicationMain(argc, argv);
}
