--!A cross-platform build utility based on Lua
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--
-- Copyright (C) 2015-present, Xmake Open Source Community.
--
-- @author      ruki
-- @file        doxygen.lua
--

task("doxygen")
    set_category("plugin")
    on_run("main")
    set_menu {
        usage = "xmake doxygen [options] [arguments]",
        description = "Generate the doxygen document.",
        options = {
            {'o', "outputdir", "kv", nil,   "Set the output directory."},
            {},
            {nil, "srcdir",    "v",  "src", "Set the source code directory."},
        }
    }
