--!A cross-platform build utility based on Lua
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--
-- Copyright (C) 2015-present, Xmake Open Source Community.
--
-- @author      ruki
-- @file        info.lua
--

-- imports
import("core.base.option")

-- get menu options
function menu_options()

    -- description
    local description = "Show information of the given packages."

    -- menu options
    local options =
    {
        {'k', "kind",       "kv", nil, "Enable static/shared library.",
                                       values = {"static", "shared"}         },
        {'p', "plat",       "kv", nil, "Set the given platform."             },
        {'a', "arch",       "kv", nil, "Set the given architecture."         },
        {'m', "mode",       "kv", nil, "Set the given mode.",
                                       values = {"release", "debug"}         },
        {'f', "configs",    "kv", nil, "Set the given extra package configs.",
                                       "e.g.",
                                       "    - xrepo fetch --configs=\"runtimes='MD'\" zlib",
                                       "    - xrepo fetch --configs=\"regex=true,thread=true\" boost"},
        {nil, "depgraph",   "k",  nil, "Show the dependency graph of the given packages.",
                                       "e.g.",
                                       "    - xrepo info --depgraph libpng",
                                       "    - xrepo info --depgraph --format=json libpng",
                                       "    - xrepo info --depgraph --format=dot libpng"},
        {nil, "format",     "kv", nil, "Set the output format.",
                                       "e.g.",
                                       "    - xrepo info --format=json zlib",
                                       "    - xrepo info --depgraph --format=dot libpng",
                                       "values: json (for --info/--depgraph), dot (for --depgraph only)",
                                       values = {"plain", "json", "dot"}},
        {},
        {nil, "packages",   "vs", nil, "The packages list.",
                                       "e.g.",
                                       "    - xrepo info zlib boost"}
    }

    -- show menu options
    local function show_options()

        -- show usage
        cprint("${bright}Usage: $${clear cyan}xrepo info [options] packages")

        -- show description
        print("")
        print(description)

        -- show options
        option.show_options(options, "info")
    end
    return options, show_options, description
end

-- info packages
function _info_packages(packages)

    -- enter working project directory
    local workdir = path.join(os.tmpdir(), "xrepo", "working")
    if not os.isdir(workdir) then
        os.mkdir(workdir)
        os.cd(workdir)
        os.vrunv(os.programfile(), {"create", "-P", "."})
    else
        os.cd(workdir)
    end

    -- do configure first, use `-q` to suppress checking noise
    -- unless `-vD` is enabled for diagnosis
    local config_argv = {"f", "-c"}
    if option.get("diagnosis") then
        table.insert(config_argv, "-vD")
    else
        table.insert(config_argv, "-q")
    end
    if option.get("plat") then
        table.insert(config_argv, "-p")
        table.insert(config_argv, option.get("plat"))
    end
    if option.get("arch") then
        table.insert(config_argv, "-a")
        table.insert(config_argv, option.get("arch"))
    end
    local mode = option.get("mode")
    if mode then
        table.insert(config_argv, "-m")
        table.insert(config_argv, mode)
    end
    local kind = option.get("kind")
    if kind then
        table.insert(config_argv, "-k")
        table.insert(config_argv, kind)
    end
    os.vrunv(os.programfile(), config_argv)

    -- show package info or dependency graph
    local require_argv = {"require"}
    if option.get("depgraph") then
        table.insert(require_argv, "--depgraph")
    else
        table.insert(require_argv, "--info")
    end
    local format = option.get("format")
    if format then
        table.insert(require_argv, "--format=" .. format)
    end
    if option.get("verbose") then
        table.insert(require_argv, "-v")
    end
    if option.get("diagnosis") then
        table.insert(require_argv, "-D")
    end
    local extra = {system = false}
    if mode == "debug" then
        extra.debug = true
    end
    if kind then
        extra.configs = extra.configs or {}
        extra.configs.shared = kind == "shared"
    end
    local configs = option.get("configs")
    if configs then
        extra.configs = extra.configs or {}
        local extra_configs, errors = ("{" .. configs .. "}"):deserialize()
        if extra_configs then
            table.join2(extra.configs, extra_configs)
        else
            raise(errors)
        end
    end
    if extra then
        local extra_str = string.serialize(extra, {indent = false, strip = true})
        table.insert(require_argv, "--extra=" .. extra_str)
    end
    table.join2(require_argv, packages)
    os.vexecv(os.programfile(), require_argv)
end

-- main entry
function main()
    local packages = option.get("packages")
    if packages then
        _info_packages(packages)
    else
        raise("please specify the packages to be viewed.")
    end
end
